@@ -1,4 +1,5 @@
 import re
+import sys
 import zlib
 from io import BytesIO
 
@@ -10,12 +11,19 @@ from .helper import (
     PillowLeakTestCase,
     assert_image,
     assert_image_equal,
+    assert_image_equal_tofile,
     hopper,
     is_big_endian,
     is_win32,
+    mark_if_feature_version,
     skip_unless_feature,
 )
 
+try:
+    import defusedxml.ElementTree as ElementTree
+except ImportError:
+    ElementTree = None
+
 # sample png stream
 
 TEST_PNG_FILE = "Tests/images/hopper.png"
@@ -106,7 +114,8 @@ class TestFilePng:
 
         test_file = "Tests/images/broken.png"
         with pytest.raises(OSError):
-            Image.open(test_file)
+            with Image.open(test_file):
+                pass
 
     def test_bad_text(self):
         # Make sure PIL can read malformed tEXt chunks (@PIL152)
@@ -324,7 +333,9 @@ class TestFilePng:
 
         with Image.open(TEST_PNG_FILE) as im:
             # Assert that there is no unclosed file warning
-            pytest.warns(None, im.verify)
+            with pytest.warns(None) as record:
+                im.verify()
+            assert not record
 
         with Image.open(TEST_PNG_FILE) as im:
             im.load()
@@ -380,25 +391,12 @@ class TestFilePng:
         # Check dpi roundtripping
 
         with Image.open(TEST_PNG_FILE) as im:
-            im = roundtrip(im, dpi=(100, 100))
-        assert im.info["dpi"] == (100, 100)
+            im = roundtrip(im, dpi=(100.33, 100.33))
+        assert im.info["dpi"] == (100.33, 100.33)
 
-    def test_load_dpi_rounding(self):
-        # Round up
+    def test_load_float_dpi(self):
         with Image.open(TEST_PNG_FILE) as im:
-            assert im.info["dpi"] == (96, 96)
-
-        # Round down
-        with Image.open("Tests/images/icc_profile_none.png") as im:
-            assert im.info["dpi"] == (72, 72)
-
-    def test_save_dpi_rounding(self):
-        with Image.open(TEST_PNG_FILE) as im:
-            im = roundtrip(im, dpi=(72.2, 72.2))
-        assert im.info["dpi"] == (72, 72)
-
-        im = roundtrip(im, dpi=(72.8, 72.8))
-        assert im.info["dpi"] == (73, 73)
+            assert im.info["dpi"] == (95.9866, 95.9866)
 
     def test_roundtrip_text(self):
         # Check text roundtripping
@@ -464,7 +462,8 @@ class TestFilePng:
 
         pngfile = BytesIO(data)
         with pytest.raises(OSError):
-            Image.open(pngfile)
+            with Image.open(pngfile):
+                pass
 
     def test_trns_rgb(self):
         # Check writing and reading of tRNS chunks for RGB images.
@@ -513,6 +512,8 @@ class TestFilePng:
 
     def test_discard_icc_profile(self):
         with Image.open("Tests/images/icc_profile.png") as im:
+            assert "icc_profile" in im.info
+
             im = roundtrip(im, icc_profile=None)
         assert "icc_profile" not in im.info
 
@@ -537,6 +538,12 @@ class TestFilePng:
             assert repr_png.format == "PNG"
             assert_image_equal(im, repr_png)
 
+    def test_repr_png_error(self):
+        im = hopper("F")
+
+        with pytest.raises(ValueError):
+            im._repr_png_()
+
     def test_chunk_order(self, tmp_path):
         with Image.open("Tests/images/icc_profile.png") as im:
             test_file = str(tmp_path / "temp.png")
@@ -565,8 +572,8 @@ class TestFilePng:
         assert len(chunks) == 3
 
     def test_read_private_chunks(self):
-        im = Image.open("Tests/images/exif.png")
-        assert im.private_chunks == [(b"orNT", b"\x01")]
+        with Image.open("Tests/images/exif.png") as im:
+            assert im.private_chunks == [(b"orNT", b"\x01")]
 
     def test_roundtrip_private_chunk(self):
         # Check private chunk roundtripping
@@ -613,6 +620,54 @@ class TestFilePng:
         with Image.open("Tests/images/hopper_idat_after_image_end.png") as im:
             assert im.text == {"TXT": "VALUE", "ZIP": "VALUE"}
 
+    def test_padded_idat(self):
+        # This image has been manually hexedited
+        # so that the IDAT chunk has padding at the end
+        # Set MAXBLOCK to the length of the actual data
+        # so that the decoder finishes reading before the chunk ends
+        MAXBLOCK = ImageFile.MAXBLOCK
+        ImageFile.MAXBLOCK = 45
+        ImageFile.LOAD_TRUNCATED_IMAGES = True
+
+        with Image.open("Tests/images/padded_idat.png") as im:
+            im.load()
+
+            ImageFile.MAXBLOCK = MAXBLOCK
+            ImageFile.LOAD_TRUNCATED_IMAGES = False
+
+            assert_image_equal_tofile(im, "Tests/images/bw_gradient.png")
+
+    def test_specify_bits(self, tmp_path):
+        im = hopper("P")
+
+        out = str(tmp_path / "temp.png")
+        im.save(out, bits=4)
+
+        with Image.open(out) as reloaded:
+            assert len(reloaded.png.im_palette[1]) == 48
+
+    def test_plte_length(self, tmp_path):
+        im = Image.new("P", (1, 1))
+        im.putpalette((1, 1, 1))
+
+        out = str(tmp_path / "temp.png")
+        im.save(str(tmp_path / "temp.png"))
+
+        with Image.open(out) as reloaded:
+            assert len(reloaded.png.im_palette[1]) == 3
+
+    def test_getxmp(self):
+        with Image.open("Tests/images/color_snakes.png") as im:
+            if ElementTree is None:
+                with pytest.warns(UserWarning):
+                    assert im.getxmp() == {}
+            else:
+                xmp = im.getxmp()
+
+                description = xmp["xmpmeta"]["RDF"]["Description"]
+                assert description["PixelXDimension"] == "10"
+                assert description["subject"]["Seq"] is None
+
     def test_exif(self):
         # With an EXIF chunk
         with Image.open("Tests/images/exif.png") as im:
@@ -648,6 +703,9 @@ class TestFilePng:
             exif = reloaded._getexif()
         assert exif[274] == 1
 
+    @mark_if_feature_version(
+        pytest.mark.valgrind_known_error, "libjpeg_turbo", "2.0", reason="Known Failing"
+    )
     def test_exif_from_jpg(self, tmp_path):
         with Image.open("Tests/images/pil_sample_rgb.jpg") as im:
             test_file = str(tmp_path / "temp.png")
@@ -676,6 +734,32 @@ class TestFilePng:
             with pytest.raises(EOFError):
                 im.seek(1)
 
+    @pytest.mark.parametrize("buffer", (True, False))
+    def test_save_stdout(self, buffer):
+        old_stdout = sys.stdout
+
+        if buffer:
+
+            class MyStdOut:
+                buffer = BytesIO()
+
+            mystdout = MyStdOut()
+        else:
+            mystdout = BytesIO()
+
+        sys.stdout = mystdout
+
+        with Image.open(TEST_PNG_FILE) as im:
+            im.save(sys.stdout, "PNG")
+
+        # Reset stdout
+        sys.stdout = old_stdout
+
+        if buffer:
+            mystdout = mystdout.buffer
+        reloaded = Image.open(mystdout)
+        assert_image_equal_tofile(reloaded, TEST_PNG_FILE)
+
 
 @pytest.mark.skipif(is_win32(), reason="Requires Unix or macOS")
 @skip_unless_feature("zlib")
